@@ -495,6 +495,21 @@ class ModelPredictiveControl:
         if np.isnan(self.costs.value).any():
             raise RuntimeError('There are still nan values in self.costs.value, something is wrong')
 
+    def reset(self):
+        """
+        Reset the underlying microgrid.
+
+        If already reset (e.g. current step is initial step), do nothing.
+
+        Returns
+        -------
+        obs : dict[str, list[float]]
+            Observations from resetting the modules as well as the flushed balance log.
+
+        """
+        if self.microgrid.current_step != self.microgrid.initial_step:
+            self.microgrid.reset()
+
     def run(self, max_steps=None, verbose=False):
         """
         Run the model prediction control algorithm.
@@ -538,13 +553,9 @@ class ModelPredictiveControl:
     def _run_mpc_on_modular(self, forecast_steps=None, verbose=False):
 
         num_iter = self._get_num_iter(forecast_steps)
-        self.microgrid.reset()
 
-        for i in tqdm(range(num_iter), desc="MPC Progress", disable=(not verbose)):
-            control = self._set_and_solve(*self._get_modular_state_values(),
-                                         iteration=i,
-                                         total_iterations=num_iter,
-                                          verbose=verbose>1)
+        for _ in tqdm(range(num_iter), desc="MPC Progress", disable=(not verbose)):
+            control = self.get_action()
 
             _, _, done, _ = self.microgrid.run(control, normalized=False)
 
@@ -553,6 +564,8 @@ class ModelPredictiveControl:
 
         return self.microgrid.get_log()
 
+    def get_action(self, verbose=0):
+        return self._set_and_solve(*self._get_modular_state_values(), verbose=verbose > 1)
 
     def _get_num_iter(self, forecast_steps=None):
         if forecast_steps is not None:
@@ -733,8 +746,6 @@ class ModelPredictiveControl:
                       cost_co2,
                       grid_co2,
                       genset_co2,
-                      iteration=None,
-                      total_iterations=None,
                       return_steps=0,
                        verbose=False):
         """
@@ -895,9 +906,9 @@ class ModelPredictiveControl:
                     f"Flattening to the difference, leading to a {'import' if grid_diff > 0 else 'export'} of {grid_diff}.")
 
         if "grid" in self.microgrid_module_names.keys():
-            control.update({self.microgrid_module_names["grid"]: grid_diff})
+            control[self.microgrid_module_names["grid"]] = [grid_diff]
 
-        control.update({self.microgrid_module_names["battery"]: battery_diff})
+        control[self.microgrid_module_names["battery"]] = [battery_diff]
 
         return control
 
