@@ -7,11 +7,13 @@ from warnings import warn
 
 from pymgrid.microgrid import DEFAULT_HORIZON
 from pymgrid.modules import ModuleContainer, UnbalancedEnergyModule
-from pymgrid.utils.logger import ModularLogger
 from pymgrid.microgrid.utils.step import MicrogridStep
+from pymgrid.utils.eq import verbose_eq
+from pymgrid.utils.logger import ModularLogger
 from pymgrid.utils.serialize import add_numpy_pandas_representers, add_numpy_pandas_constructors, dump_data
 from pymgrid.utils.space import MicrogridSpace
 
+
 class Microgrid(yaml.YAMLObject):
     """
     Microgrid class, used to define and simulate an environment with a variety of modules.
@@ -111,10 +113,9 @@ class Microgrid(yaml.YAMLObject):
                                                    overgeneration_cost)
 
         # TODO (ahalev) transform envs to wrappers, and remove microgrid from attr names)
-        self.microgrid_action_space = MicrogridSpace(
-            self._modules.get_attrs('action_space', 'module_type', as_pandas=False), 'act'
-        )
-        self.microgrid_observation_space = MicrogridSpace(
+        self.microgrid_action_space = MicrogridSpace.from_module_spaces(
+            self._modules.get_attrs('action_space', 'module_type', as_pandas=False), 'act')
+        self.microgrid_observation_space = MicrogridSpace.from_module_spaces(
             self._modules.get_attrs('observation_space', as_pandas=False), 'obs'
         )
 
@@ -477,12 +478,32 @@ class Microgrid(yaml.YAMLObject):
         for key, value in self._balance_logger.to_dict().items():
             _log_dict[('balance', 0, key)] = value
 
+        pad = (0, '')
+
         for key, value in self._microgrid_logger.items():
-            _log_dict[(key, 0, '')] = value
+            key = key if pd.api.types.is_list_like(key) else [key]
+            _log_dict[(*key, *pad[len(key)-1:])] = value
 
         col_names = ['module_name', 'module_number', 'field']
 
-        df = pd.DataFrame(_log_dict, index=pd.RangeIndex(start=self.initial_step, stop=self.current_step))
+        initial_step = self._modules.get_attrs('initial_step', unique=True)
+
+        try:
+            df = pd.DataFrame(_log_dict, index=pd.RangeIndex(start=initial_step, stop=self.current_step))
+        except ValueError as e:
+            if 'Length of values' in e.args[0]:
+                module_log_lengths = pd.Series([len(log_dict) for log_dict in _log_dict.values()])
+
+                msg = f"Length of module log dicts ({module_log_lengths.unique().item()}) " \
+                      f"do not match self.current_step-initial_step ({self.current_step-initial_step}). " \
+                      f"Did you set a trajectory attribute " \
+                      f"('initial_step', 'final_step', 'trajectory_func') without calling Microgrid.reset()?"
+
+                raise ValueError(msg)
+
+            else:
+                raise
+
         df.columns = pd.MultiIndex.from_tuples(df.columns.to_list(), names=col_names)
 
         if drop_forecasts:
@@ -639,7 +660,7 @@ class Microgrid(yaml.YAMLObject):
         current_step : int
             Current step.
         """
-        return self._modules.get_attrs('current_step', unique=True).item()
+        return self._modules.get_attrs('current_step', unique=True)
 
     @property
     def initial_step(self):
@@ -656,12 +677,10 @@ class Microgrid(yaml.YAMLObject):
         return self._initial_step
 
     def _get_module_initial_step(self):
-        initial_step = self.modules.get_attrs('initial_step', unique=True)
         try:
-            return initial_step.item()
-        except ValueError:
-            if initial_step.empty:
-                return 0
+            return self.modules.get_attrs('initial_step', unique=True)
+        except AttributeError:
+            return 0
 
     @initial_step.setter
     def initial_step(self, value):
@@ -685,12 +704,10 @@ class Microgrid(yaml.YAMLObject):
         return self._final_step
 
     def _get_module_final_step(self):
-        final_step = self.modules.get_attrs('final_step', unique=True)
         try:
-            return final_step.item()
-        except ValueError:
-            if final_step.empty:
-                return np.inf
+            return self.modules.get_attrs('final_step', unique=True)
+        except AttributeError:
+            return np.inf
 
     @final_step.setter
     def final_step(self, value):
@@ -714,7 +731,7 @@ class Microgrid(yaml.YAMLObject):
         """
         return self._modules
 
-    def state_dict(self, normalized=False):
+    def state_dict(self, normalized=False, as_run_output=False):
         """
         State of the microgrid as a dict.
 
@@ -725,14 +742,24 @@ class Microgrid(yaml.YAMLObject):
         normalized : bool, default False
             Whether to return a dict of normalized values.
 
+        as_run_output : bool, default False
+            Whether to return output in the same format as the output of :meth:`Microgrid.run`.
+            Inner values are numpy arrays and not dict in this case.
+
+
         Returns
         -------
-        state_dict : dict[str, list[dict]]
+        state_dict : dict[str, list[dict]] or dict[str[list[np.ndarray]]]
             State of the microgrid as a nested dict.
 
         """
+        def as_run_output_f(state_dict):
+            if as_run_output:
+                return np.array(list(state_dict.values()))
+            return state_dict
+
         return {name: [
-            module.state_dict(normalized=normalized) for module in modules
+            as_run_output_f(module.state_dict(normalized=normalized)) for module in modules
         ] for name, modules in self._modules.iterdict()}
 
     @property
@@ -1007,6 +1034,9 @@ class Microgrid(yaml.YAMLObject):
             not x.startswith('_') and not callable(getattr(self._modules, x)) and x in self._modules
         }
 
+    def verbose_eq(self, other, indent=0):
+        verbose_eq(self, other, ('_balance_logger', 'trajectory_func', *self._modules.names()), indent=indent)
+
     def __dir__(self):
         rv = set(super().__dir__())
         rv = rv | self._dir_additions()
