@@ -6,6 +6,7 @@ import numpy as np
 
 from warnings import warn
 
+from pymgrid.utils.eq import verbose_eq
 from pymgrid.utils.logger import ModularLogger
 from pymgrid.utils.space import ModuleSpace
 from pymgrid.utils.serialize import add_numpy_pandas_representers, add_numpy_pandas_constructors, dump_data
@@ -37,6 +38,7 @@ class BaseMicrogridModule(yaml.YAMLObject):
     def __init__(self,
                  raise_errors,
                  initial_step=0,
+                 normalized_action_bounds=(0, 1),
                  provided_energy_name='provided_energy',
                  absorbed_energy_name='absorbed_energy'
                  ):
@@ -44,17 +46,18 @@ class BaseMicrogridModule(yaml.YAMLObject):
         self.raise_errors = raise_errors
         self.initial_step = initial_step
         self._current_step = initial_step
-        self._action_space = self._get_action_spaces()
+        self._action_space = self._get_action_spaces(normalized_action_bounds)
         self._observation_space = self._get_observation_spaces()
         self.provided_energy_name, self.absorbed_energy_name = provided_energy_name, absorbed_energy_name
         self._logger = ModularLogger()
-        self.name = (None, None)  # set by ModularMicrogrid
+        self.name = None  # set by ModularMicrogrid
 
-    def _get_action_spaces(self):
+    def _get_action_spaces(self, normalized_bounds):
         unnormalized_low = self.min_act if isinstance(self.min_act, np.ndarray) else np.array([self.min_act])
         unnormalized_high = self.max_act if isinstance(self.max_act, np.ndarray) else np.array([self.max_act])
         return ModuleSpace(unnormalized_low=unnormalized_low,
-                           unnormalized_high=unnormalized_high)
+                           unnormalized_high=unnormalized_high,
+                           normalized_bounds=normalized_bounds)
 
     def _get_observation_spaces(self):
         unnormalized_low = self.min_obs if isinstance(self.min_obs, np.ndarray) else np.array([self.min_obs])
@@ -203,9 +206,6 @@ class BaseMicrogridModule(yaml.YAMLObject):
         """
 
         assert energy_demand >= 0
-        assert self.is_source, f'step() was called with positive energy (source) for module {self} but ' \
-                               f'module is not a source and ' \
-                               f'can only be called with negative energy.'
 
         if self.module_type[-1] == 'fixed':
             return self.update(None, as_source=True)
@@ -302,8 +302,8 @@ class BaseMicrogridModule(yaml.YAMLObject):
 
         Parameters
         ----------
-        external_energy_change : float or None
-            Amount of energy to provide or absorb.
+        external_energy_change : float >= 0 or None
+            Amount of energy to provide or absorb. A non-negative number.
         as_source : bool
             Whether the module is acting as a source.
         as_sink
@@ -405,6 +405,25 @@ class BaseMicrogridModule(yaml.YAMLObject):
         if obs:
             return self._observation_space.denormalize(value)
 
+    def dynamic_action_space(self):
+        """
+        An action space bounded by the current step's maximum consumption and production.
+
+        This is useful for checking if an action satisfies production or consumption limits in a given step.
+        It can also be used for clipping actions to these bounds by using :meth:`.ModuleSpace.clip`.
+
+        Returns
+        -------
+
+        action_space : :class:`.ModuleSpace`
+            A space with bounds for the current step.
+
+        """
+        return ModuleSpace(unnormalized_low=-1*self.max_consumption,
+                           unnormalized_high=self.max_production,
+                           normalized_bounds=self.normalized_action_bounds,
+                           shape=self._action_space.shape)
+
     def log_dict(self):
         """
         Module's log as a dict.
@@ -624,14 +643,13 @@ class BaseMicrogridModule(yaml.YAMLObject):
         Maximum amount of production at the current time step.
 
         Must be defined in any child class that is a source.
-        If the module is not a source, this value is irrelevant.
 
         Returns
         -------
         float
 
         """
-        return NotImplemented
+        return 0.0
 
     @property
     def max_consumption(self):
@@ -639,14 +657,13 @@ class BaseMicrogridModule(yaml.YAMLObject):
         Maximum amount of consumption at the current time step.
 
         Must be defined in any child class that is a sink.
-        If the module is not a sink, this value is irrelevant.
 
         Returns
         -------
         float
 
         """
-        return NotImplemented
+        return 0.0
 
     @property
     def marginal_cost(self):
@@ -697,6 +714,49 @@ class BaseMicrogridModule(yaml.YAMLObject):
         """
         return self._observation_space
 
+    @property
+    def normalized_action_bounds(self):
+        """
+        Bounds that actions are normalized to in the action space.
+
+        This property is necessary for serialization.
+
+        :meta private:
+
+        Returns
+        -------
+        normalized_action_bounds : tuple
+            Normalized bounds
+        """
+
+        low = np.unique(self._action_space.normalized.low)
+        high = np.unique(self._action_space.normalized.high)
+
+        try:
+            low = low.item()
+            high = high.item()
+        except ValueError:
+            if not self._action_space.shape == (0, ):
+                raise
+            low, high = 0, 1
+
+        if isinstance(low, float) and low.is_integer():
+            low = int(low)
+
+        if isinstance(high, float) and high.is_integer():
+            high = int(high)
+
+        return low, high
+
+    @normalized_action_bounds.setter
+    def normalized_action_bounds(self, value):
+        if len(value) != 2:
+            raise TypeError('Invalid number of components.')
+        if value[0] > value[1]:
+            raise ValueError('Low value cannot be greater than high value.')
+
+        self._action_space = self._get_action_spaces(value)
+
     @property
     def is_source(self):
         """
@@ -956,12 +1016,30 @@ class BaseMicrogridModule(yaml.YAMLObject):
             warn(f"Unused keys in serialized_dict: {list(serialized_dict.keys())}")
         return self
 
+    def verbose_eq(self, other, indent=0):
+        return verbose_eq(self, other, self.__dict__.keys(), indent=indent)
+
     def __eq__(self, other):
         if type(self) != type(other):
             return NotImplemented
 
-        diff = [(k1, v1, v2) for (k1, v1), (k2, v2) in zip(self.__dict__.items(), other.__dict__.items()) if
-                ((hasattr(v1, "any") and not np.allclose(v1, v2)) or (not hasattr(v1, "any") and v1 != v2))]
+        def are_equal(v1, v2):
+            try:
+                _are_equal = bool(v1 == v2)
+                if _are_equal:
+                    return True
+            except ValueError:
+                pass
+
+            try:
+                return np.allclose(v1, v2)
+            except (ValueError, TypeError):
+                return False
+
+        diff = [
+            (k1, v1, v2) for (k1, v1), (k2, v2) in zip(self.__dict__.items(), other.__dict__.items())
+            if not are_equal(v1, v2)
+        ]
 
         return len(diff) == 0
 
