@@ -69,6 +69,7 @@ class GensetModule(BaseMicrogridModule):
                  allow_abortion=True,
                  init_start_up=True,
                  initial_step=0,
+                 normalized_action_bounds=(0, 1),
                  raise_errors=False,
                  provided_energy_name='genset_production'):
 
@@ -90,10 +91,10 @@ class GensetModule(BaseMicrogridModule):
 
         self._current_status, self._goal_status = int(init_start_up), int(init_start_up)
         self._steps_until_up, self._steps_until_down = self._reset_up_down_times()
-        self.name = ('genset', None)
 
         super().__init__(raise_errors,
                          initial_step=initial_step,
+                         normalized_action_bounds=normalized_action_bounds,
                          provided_energy_name=provided_energy_name,
                          absorbed_energy_name=None)
 
@@ -143,10 +144,16 @@ class GensetModule(BaseMicrogridModule):
             this module provided to or absorbed from the microgrid.
 
         """
-        goal_status = action[0]
+        if normalized:
+            denormalized = self._action_space.denormalize(action)
+        else:
+            denormalized = action
+
+        goal_status = denormalized[0]
+
         assert 0 <= goal_status <= 1
         self.update_status(goal_status)
-        return super().step(action, normalized=normalized)
+        return super().step(denormalized, normalized=False)
 
     def get_co2(self, production):
         """
@@ -205,7 +212,9 @@ class GensetModule(BaseMicrogridModule):
         return self._get_fuel_cost(production) + self.get_co2_cost(production)
 
     def update(self, external_energy_change, as_source=False, as_sink=False):
-        assert as_source, 'This module may only act as a source.'
+        assert as_source or external_energy_change == 0.0, f'step() was called with negative energy (sink) for ' \
+                                                           f'module {self} but module is not a sink and ' \
+                                                           f'can only be called with positive energy.'
 
         reward = -1.0 * self.get_cost(external_energy_change)
         info = {'provided_energy': external_energy_change,
