@@ -15,15 +15,23 @@ class ModularLogger(UserDict):
         self._log_length = 0
         return d
 
-    def log(self, **log_dict):
+    def log(self, log_dict=None, **log_items):
+        if log_items:
+            if log_dict:
+                raise TypeError('Cannot pass both positional and keyword arguments.')
+
+            log_dict = log_items
+
         for key, value in log_dict.items():
+            if key not in self:
+                self[key] = []
+
             try:
                 self[key].append(value.item())
             except AttributeError:
                 self[key].append(value)
-            except KeyError:
-                self[key] = [np.nan]*self._log_length
-                self[key].append(value)
+            except ValueError:
+                raise ValueError('Only scalar values can be logged.')
 
         self._log_length += 1
 
